@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # This file is part of REANA.
-# Copyright (C) 2018, 2019, 2020, 2021 CERN.
+# Copyright (C) 2018, 2019, 2020, 2021, 2022, 2023 CERN.
 #
 # REANA is free software; you can redistribute it and/or modify it
 # under the terms of the MIT License; see LICENSE file for more details.
@@ -16,12 +16,15 @@ import logging
 import uuid
 from datetime import datetime
 from functools import reduce
+from typing import Dict, List
 
 from reana_commons.config import (
     MQ_MAX_PRIORITY,
     REANA_MAX_CONCURRENT_BATCH_WORKFLOWS,
     REANA_RUNTIME_KUBERNETES_KEEP_ALIVE_JOBS_WITH_STATUSES,
+    WORKFLOW_TIME_FORMAT,
 )
+from reana_commons.errors import REANAValidationError
 from reana_commons.utils import get_disk_usage
 from sqlalchemy import (
     BigInteger,
@@ -31,6 +34,7 @@ from sqlalchemy import (
     Enum,
     Float,
     ForeignKey,
+    Integer,
     String,
     Text,
     UniqueConstraint,
@@ -50,14 +54,15 @@ from reana_db.config import (
     DB_SECRET_KEY,
     DEFAULT_QUOTA_LIMITS,
     DEFAULT_QUOTA_RESOURCES,
+    LIMIT_RESTARTS,
     WORKFLOW_TERMINATION_QUOTA_UPDATE_POLICY,
-    QuotaResourceType,
 )
 from reana_db.utils import (
     build_workspace_path,
     store_workflow_disk_quota,
-    get_default_quota_resource,
     update_users_disk_quota,
+    update_workflow_cpu_quota,
+    update_workspace_retention_rules,
 )
 
 
@@ -80,8 +85,8 @@ class QuotaBase:
             health = QuotaHealth.healthy
             if limit:
                 percentage = usage / limit * 100
-                if percentage >= 60:
-                    if percentage >= 85:
+                if percentage >= 80:
+                    if percentage >= 100:
                         health = QuotaHealth.critical
                     else:
                         health = QuotaHealth.warning
@@ -357,6 +362,7 @@ ALLOWED_WORKFLOW_STATUS_TRANSITIONS = [
     # Queued
     (RunStatus.queued, RunStatus.deleted),
     (RunStatus.queued, RunStatus.pending),
+    (RunStatus.queued, RunStatus.failed),
     # Pending
     (RunStatus.pending, RunStatus.running),
     (RunStatus.pending, RunStatus.deleted),
@@ -465,8 +471,8 @@ class Workflow(Base, Timestamp, QuotaBase):
     git_ref = Column(String(40))
     git_repo = Column(String(255))
     git_provider = Column(String(255))
+    launcher_url = Column(String)
     owner = relationship("User", backref="workflow")
-
     sessions = relationship(
         "InteractiveSession",
         secondary="__reana.workflow_session",
@@ -474,6 +480,9 @@ class Workflow(Base, Timestamp, QuotaBase):
         backref="workflow",
         cascade="all, delete",
     )
+    retention_rules = relationship(
+        "WorkspaceRetentionRule", backref="workflow", lazy="dynamic"
+    )
 
     __table_args__ = (
         UniqueConstraint(
@@ -500,6 +509,7 @@ class Workflow(Base, Timestamp, QuotaBase):
         workspace_path=None,
         restart=False,
         run_number=None,
+        launcher_url=None,
     ):
         """Initialize workflow model."""
         self.id_ = id_
@@ -520,6 +530,7 @@ class Workflow(Base, Timestamp, QuotaBase):
         self.workspace_path = workspace_path or build_workspace_path(
             self.owner_id, self.id_
         )
+        self.launcher_url = launcher_url
 
     def __repr__(self):
         """Workflow string representation."""
@@ -560,6 +571,13 @@ class Workflow(Base, Timestamp, QuotaBase):
                 .first()
             )
         if last_workflow and self.restart:
+            # FIXME: remove the limit of nine restarts when we fix the way in which
+            # we save `run_number` in the DB
+            num_restarts = round(last_workflow.run_number * 10) % 10
+            if num_restarts == LIMIT_RESTARTS:
+                raise REANAValidationError(
+                    f"Cannot restart a workflow more than {LIMIT_RESTARTS} times"
+                )
             return round(last_workflow.run_number + 0.1, 1)
         else:
             if not last_workflow:
@@ -591,29 +609,75 @@ class Workflow(Base, Timestamp, QuotaBase):
         """Retrieve disk usage information of a workspace."""
         from functools import partial
 
-        if not summarize:
-            # size break down per directory so we can't query DB (`r-client du`)
-            return get_disk_usage(
-                self.workspace_path,
-                summarize,
-                search,
-                to_human_readable_units=partial(
-                    ResourceUnit.human_readable_unit, ResourceUnit.bytes_
-                ),
-            )
+        return get_disk_usage(
+            self.workspace_path,
+            summarize,
+            search,
+            to_human_readable_units=partial(
+                ResourceUnit.human_readable_unit, ResourceUnit.bytes_
+            ),
+        )
 
-        disk_usage = self.get_quota_usage().get("disk", {}).get("usage", {})
-        if not disk_usage:
-            # recalculate disk workflow resource
-            workflow_resource = store_workflow_disk_quota(self)
-            disk_usage = dict(
-                raw=workflow_resource.quota_used,
-                to_human_readable_units=ResourceUnit.human_readable_unit(
-                    ResourceUnit.bytes_, workflow_resource.quota_used
-                ),
-            )
+    def set_workspace_retention_rules(self, rules: List[Dict[str, str]]):
+        """Set workspace retention rules for the workflow."""
+        from .database import Session
 
-        return [{"name": "", "size": disk_usage}]
+        for rule in rules:
+            wr = WorkspaceRetentionRule(
+                workflow_id=self.id_,
+                workspace_files=rule["workspace_files"],
+                retention_days=rule["retention_days"],
+            )
+            Session.add(wr)
+        Session.commit()
+
+    def activate_workspace_retention_rules(self):
+        """Activate workspace retention rules for the workflow."""
+        rules = WorkspaceRetentionRule.query.filter_by(
+            workflow_id=self.id_, status=WorkspaceRetentionRuleStatus.created
+        )
+        update_workspace_retention_rules(rules, WorkspaceRetentionRuleStatus.active)
+
+    def get_all_restarts(self):
+        """Get all the restarts of this workflow, including the original workflow.
+
+        Returns all the restarts of this workflow, that is all the workflows that have
+        the same name and the same run number (up to the dot). This includes the
+        original workflow, as well as all the following restarts.
+        """
+        run_number = int(self.run_number)
+        restarts = Workflow.query.filter(
+            Workflow.name == self.name,
+            Workflow.owner_id == self.owner_id,
+            Workflow.run_number >= run_number,
+            Workflow.run_number < run_number + 1,
+        )
+        return restarts
+
+    def inactivate_workspace_retention_rules(self):
+        """Inactivate workspace retention rules for all the parent workflows."""
+        rules = WorkspaceRetentionRule.query.join(
+            self.get_all_restarts().subquery()
+        ).filter(
+            or_(
+                WorkspaceRetentionRule.status == WorkspaceRetentionRuleStatus.created,
+                WorkspaceRetentionRule.status == WorkspaceRetentionRuleStatus.active,
+            ),
+        )
+        update_workspace_retention_rules(rules, WorkspaceRetentionRuleStatus.inactive)
+
+    def workspace_has_pending_retention_rules(self):
+        """Check whether the workspace has retention rules that are pending.
+
+        All the restarts of the workflow are considered when checking the retention
+        rules, as they all share the same workspace.
+        """
+        pending_rules = WorkspaceRetentionRule.query.join(
+            self.get_all_restarts().subquery()
+        ).filter(
+            WorkspaceRetentionRule.status == WorkspaceRetentionRuleStatus.pending,
+        )
+        return pending_rules.first() is not None
 
     def get_priority(self, cluster_memory):
         """Workflow priority when scheduling it.
@@ -687,36 +751,29 @@ class Workflow(Base, Timestamp, QuotaBase):
 @event.listens_for(Workflow.status, "set")
 def workflow_status_change_listener(workflow, new_status, old_status, initiator):
     """Workflow status change listener."""
-    from .database import Session
 
     def _update_disk_quota(workflow):
-        if QuotaResourceType.disk not in WORKFLOW_TERMINATION_QUOTA_UPDATE_POLICY:
+        if ResourceType.disk.name not in WORKFLOW_TERMINATION_QUOTA_UPDATE_POLICY:
             return
-        update_users_disk_quota(user=workflow.owner)
-        store_workflow_disk_quota(workflow)
+
+        try:
+            update_users_disk_quota(user=workflow.owner)
+            store_workflow_disk_quota(workflow)
+        except Exception as e:
+            logging.error(f"Failed to update disk quota: \n{e}\nContinuing...")
 
     def _update_cpu_quota(workflow):
-        if QuotaResourceType.cpu not in WORKFLOW_TERMINATION_QUOTA_UPDATE_POLICY:
+        if ResourceType.cpu.name not in WORKFLOW_TERMINATION_QUOTA_UPDATE_POLICY:
             return
 
-        terminated_at = workflow.run_finished_at or workflow.run_stopped_at
-        if workflow.run_started_at and terminated_at:
-            cpu_time = terminated_at - workflow.run_started_at
-            cpu_milliseconds = int(cpu_time.total_seconds() * 1000)
-            cpu_resource = get_default_quota_resource(ResourceType.cpu.name)
-            workflow_resource = WorkflowResource(
-                workflow_id=workflow.id_,
-                resource_id=cpu_resource.id_,
-                quota_used=cpu_milliseconds,
-            )
-            user_resource_quota = UserResource.query.filter_by(
-                user_id=workflow.owner_id, resource_id=cpu_resource.id_
-            ).first()
-            user_resource_quota.quota_used += cpu_milliseconds
-            Session.add(workflow_resource)
-            Session.commit()
+        try:
+            update_workflow_cpu_quota(workflow=workflow)
+        except Exception as e:
+            logging.error(f"Failed to update cpu quota: \n{e}\nContinuing...")
 
     workflow.update_workflow_timestamp(new_status)
+    if new_status in [RunStatus.finished, RunStatus.failed]:
+        workflow.activate_workspace_retention_rules()
     if new_status in [
         RunStatus.finished,
         RunStatus.failed,
@@ -809,6 +866,122 @@ class AuditLog(Base, Timestamp):
         return "<AuditLog {} {}>".format(self.id_, self.action)
 
 
+class WorkspaceRetentionRuleStatus(enum.Enum):
+    """Enumeration of workspace retention rule status.
+
+    - ``created`` : initial status of each rule.
+    - ``active``  : the workflow has finished running and the rule can now be considered.
+    - ``inactive``: the rule will not be considered, even though it wasn't applied.
+    - ``pending`` : the rule is currently being handled by the cronjob.
+    - ``applied`` : the rule was handled by the cronjob and the files were deleted.
+    """
+
+    created = 0
+    active = 1
+    inactive = 2
+    applied = 3
+    pending = 4
+
+
+ALLOWED_WORKSPACE_RETENTION_RULE_STATUS_TRANSITIONS = [
+    # Created
+    (WorkspaceRetentionRuleStatus.created, WorkspaceRetentionRuleStatus.active),
+    (WorkspaceRetentionRuleStatus.created, WorkspaceRetentionRuleStatus.inactive),
+    # Active
+    (WorkspaceRetentionRuleStatus.active, WorkspaceRetentionRuleStatus.inactive),
+    (WorkspaceRetentionRuleStatus.active, WorkspaceRetentionRuleStatus.pending),
+    # Inactive
+    (WorkspaceRetentionRuleStatus.inactive, WorkspaceRetentionRuleStatus.active),
+    # Pending
+    (WorkspaceRetentionRuleStatus.pending, WorkspaceRetentionRuleStatus.applied),
+    (WorkspaceRetentionRuleStatus.pending, WorkspaceRetentionRuleStatus.active),
+    # Applied
+    (WorkspaceRetentionRuleStatus.applied, WorkspaceRetentionRuleStatus.applied),
+]
+
+
+class WorkspaceRetentionAuditLog(Base):
+    """Workspace retention audit log table."""
+
+    __tablename__ = "workspace_retention_audit_log"
+    __table_args__ = {"schema": "__reana"}
+
+    workspace_retention_rule_id = Column(
+        UUIDType,
+        ForeignKey("__reana.workspace_retention_rule.id_"),
+        primary_key=True,
+        nullable=False,
+    )
+    timestamp = Column(DateTime, server_default=func.now(), primary_key=True)
+    action = Column(
+        Enum(WorkspaceRetentionRuleStatus), primary_key=True, nullable=False
+    )
+
+    def __repr__(self):
+        """Workspace Retention audit log string representation."""
+        return f"<WorkspaceRetentionAuditLog {self.workspace_retention_rule_id} {self.action}>"
+
+
+class WorkspaceRetentionRule(Base):
+    """Workspace retention rule table."""
+
+    __tablename__ = "workspace_retention_rule"
+    __table_args__ = {"schema": "__reana"}
+
+    id_ = Column(UUIDType, primary_key=True, default=generate_uuid)
+    workflow_id = Column(UUIDType, ForeignKey("__reana.workflow.id_"), nullable=False)
+    workspace_files = Column(String(length=255), nullable=False)
+    retention_days = Column(Integer, nullable=False)
+    apply_on = Column(DateTime, nullable=True)
+    status = Column(
+        Enum(WorkspaceRetentionRuleStatus),
+        default=WorkspaceRetentionRuleStatus.created,
+        nullable=False,
+    )
+    audit_logs = relationship(
+        "WorkspaceRetentionAuditLog", backref="workspace_retention_rule", lazy="dynamic"
+    )
+
+    __table_args__ = (
+        UniqueConstraint(
+            "workflow_id", "workspace_files", name="_workspace_retention_rule_uc"
+        ),
+        {"schema": "__reana"},
+    )
+
+    def can_transition_to(self, next_status):
+        """Whether the provided retention rule can transition to the next status."""
+        current_transition = (self.status, next_status)
+        return current_transition in ALLOWED_WORKSPACE_RETENTION_RULE_STATUS_TRANSITIONS
+
+    def serialize(self):
+        """Serialize workspace retention rule object data."""
+        return {
+            "id": str(self.id_),
+            "workspace_files": self.workspace_files,
+            "retention_days": self.retention_days,
+            "apply_on": (
+                self.apply_on.strftime(WORKFLOW_TIME_FORMAT) if self.apply_on else None
+            ),
+            "status": self.status.name,
+        }
+
+    def __repr__(self):
+        """Workspace Retention string representation."""
+        return f"<WorkspaceRetentionRule {self.workflow_id} {self.workspace_files} {self.retention_days} {self.status}>"
+
+
+@event.listens_for(WorkspaceRetentionRule, "after_insert")
+@event.listens_for(WorkspaceRetentionRule, "after_update")
+def workspace_retention_change_listener(mapper, connection, workspace_retention_rule):
+    """Workspace retention change listener."""
+    connection.execute(
+        WorkspaceRetentionAuditLog.__table__.insert(),
+        workspace_retention_rule_id=workspace_retention_rule.id_,
+        action=workspace_retention_rule.status,
+    )
+
+
 class ResourceType(enum.Enum):
     """Enumeration of resource types."""
 
@@ -825,13 +998,14 @@ class ResourceUnit(enum.Enum):
     @staticmethod
     def _human_readable_milliseconds(milliseconds):
         """Convert milliseconds usage to human readable string."""
+        abs_seconds = abs(milliseconds) // 1000
         hours, minutes, seconds = (
-            milliseconds // (1000 * 60 * 60),
-            (milliseconds // (1000 * 60)) % 60,
-            (milliseconds // 1000) % 60,
+            abs_seconds // (60 * 60),
+            (abs_seconds // 60) % 60,
+            abs_seconds % 60,
         )
 
-        human_readable_milliseconds = ""
+        human_readable_milliseconds = "-" if milliseconds < 0 else ""
         for value, unit in [(hours, "h"), (minutes, "m"), (seconds, "s")]:
             if value >= 1:
                 human_readable_milliseconds += "{value}{unit} ".format(
@@ -848,7 +1022,7 @@ class ResourceUnit(enum.Enum):
         units = ["Bytes", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"]
         digits = 2
         k = 1024
-        unit_index = int(math.floor(math.log(bytes_) / math.log(k)))
+        unit_index = math.floor(math.log(abs(bytes_), k))
 
         converted_value = round(bytes_ / math.pow(k, unit_index), digits)
         return "{converted_value} {converted_unit}".format(
