@@ -1,22 +1,26 @@
 # -*- coding: utf-8 -*-
 #
 # This file is part of REANA.
-# Copyright (C) 2018, 2019, 2020, 2021 CERN.
+# Copyright (C) 2018, 2019, 2020, 2021, 2022, 2023 CERN.
 #
 # REANA is free software; you can redistribute it and/or modify it
 # under the terms of the MIT License; see LICENSE file for more details.
 """REANA-DB utils."""
 
+import logging
 import os
+from datetime import datetime, timedelta
 from typing import Optional
 from uuid import UUID
 
 from sqlalchemy import inspect
+from sqlalchemy.orm import defer
+from reana_commons.utils import get_disk_usage
+from reana_commons.errors import REANAMissingWorkspaceError
 
 from reana_db.config import (
-    CRONJOB_DISK_QUOTA_UPDATE_POLICY,
+    PERIODIC_RESOURCE_QUOTA_UPDATE_POLICY,
     WORKFLOW_TERMINATION_QUOTA_UPDATE_POLICY,
-    QuotaResourceType,
 )
 
 
@@ -62,6 +66,7 @@ def _get_workflow_with_uuid_or_name(uuid_or_name, user_uuid):
         workflow with latest run number of all the workflows with this name
         is returned.
     :type uuid_or_name: String
+    :param user_uuid: UUID of the workflow's owner.
 
     :rtype: reana-db.models.Workflow
     """
@@ -88,7 +93,7 @@ def _get_workflow_with_uuid_or_name(uuid_or_name, user_uuid):
     if is_uuid:
         # `uuid_or_name` is an UUIDv4.
         # Search with it since it is expected to be unique.
-        return _get_workflow_by_uuid(uuid_or_name)
+        return _get_workflow_by_uuid(uuid_or_name, user_uuid)
 
     else:
         # `uuid_or_name` is not and UUIDv4. Expect it is a name.
@@ -177,17 +182,20 @@ def _get_workflow_by_name(workflow_name, user_uuid):
     return workflow
 
 
-def _get_workflow_by_uuid(workflow_uuid):
+def _get_workflow_by_uuid(workflow_uuid, user_uuid):
     """Get Workflow with UUIDv4.
 
     :param workflow_uuid: UUIDv4 of a Workflow.
     :type workflow_uuid: String representing a valid UUIDv4.
+    :param user_uuid: UUID of the workflow's owner.
 
     :rtype: reana-db.models.Workflow
     """
     from reana_db.models import Workflow
 
-    workflow = Workflow.query.filter(Workflow.id_ == workflow_uuid).first()
+    workflow = Workflow.query.filter(
+        Workflow.id_ == workflow_uuid, Workflow.owner_id == user_uuid
+    ).first()
     if not workflow:
         raise ValueError(
             "REANA_WORKON is set to {0}, but "
@@ -216,7 +224,20 @@ def get_default_quota_resource(resource_type):
     return Resource.query.filter_by(name=DEFAULT_QUOTA_RESOURCES[resource_type]).one()
 
 
-def update_users_disk_quota(user=None, bytes_to_sum: Optional[int] = None) -> None:
+def should_skip_quota_update(resource_type) -> bool:
+    """Check if quota updates should be skipped based on the update policy.
+
+    :param resource_type: Resource type of the quota that needs to be updated.
+    """
+    return (
+        resource_type.name not in WORKFLOW_TERMINATION_QUOTA_UPDATE_POLICY
+        and not PERIODIC_RESOURCE_QUOTA_UPDATE_POLICY
+    )
+
+
+def update_users_disk_quota(
+    user=None, bytes_to_sum: Optional[int] = None, override_policy_checks: bool = False
+) -> None:
     """Update users disk quota usage.
 
     :param user: User whose disk quota will be updated. If None, applies to all users.
@@ -226,15 +247,14 @@ def update_users_disk_quota(user=None, bytes_to_sum: Optional[int] = None) -> No
     :type user: reana_db.models.User
     :type bytes_to_sum: int
 
+    :param override_policy_checks: Whether to update the disk quota without checking the
+        update policy.
     """
-    if (
-        QuotaResourceType.disk not in WORKFLOW_TERMINATION_QUOTA_UPDATE_POLICY
-        and not CRONJOB_DISK_QUOTA_UPDATE_POLICY
-    ):
-        return
-
     from reana_db.config import DEFAULT_QUOTA_RESOURCES
-    from reana_db.models import Resource, User, UserResource
+    from reana_db.models import Resource, ResourceType, User, UserResource
+
+    if not override_policy_checks and should_skip_quota_update(ResourceType.disk):
+        return
 
     users = [user] if user else User.query.all()
 
@@ -249,8 +269,17 @@ def update_users_disk_quota(user=None, bytes_to_sum: Optional[int] = None) -> No
             user_resource_quota = UserResource.query.filter_by(
                 user_id=u.id_, resource_id=disk_resource.id_
             ).first()
-            if bytes_to_sum:
-                user_resource_quota.quota_used += bytes_to_sum
+            if bytes_to_sum is not None:
+                updated_quota_usage = user_resource_quota.quota_used + bytes_to_sum
+                if updated_quota_usage < 0:
+                    logging.warning(
+                        f"Disk quota consumption of user {u.id_} would become negative: "
+                        f"{user_resource_quota.quota_used} [original usage] + {bytes_to_sum} [delta] "
+                        f"-> {updated_quota_usage} [new usage]. Setting the new usage to zero."
+                    )
+                    user_resource_quota.quota_used = 0
+                else:
+                    user_resource_quota.quota_used = updated_quota_usage
             else:
                 workspace_path = u.get_user_workspace()
                 disk_usage_bytes = get_disk_usage_or_zero(workspace_path)
@@ -258,17 +287,133 @@ def update_users_disk_quota(user=None, bytes_to_sum: Optional[int] = None) -> No
             Session.commit()
 
 
+def update_workflow_cpu_quota(workflow) -> int:
+    """Update workflow CPU quota based on started and finished/stopped times.
+
+    :return: Workflow running time in milliseconds if workflow has terminated, else 0.
+    """
+    from reana_db.database import Session
+
+    from reana_db.models import (
+        ResourceType,
+        UserResource,
+        WorkflowResource,
+    )
+
+    terminated_at = workflow.run_finished_at or workflow.run_stopped_at
+    if workflow.run_started_at and terminated_at:
+        cpu_time = terminated_at - workflow.run_started_at
+        cpu_milliseconds = int(cpu_time.total_seconds() * 1000)
+        cpu_resource = get_default_quota_resource(ResourceType.cpu.name)
+        # WorkflowResource might exist already if the cluster
+        # follows a combined termination + periodic policy (eg. created
+        # by the status listener, revisited by the cronjob)
+        workflow_resource = WorkflowResource.query.filter_by(
+            workflow_id=workflow.id_, resource_id=cpu_resource.id_
+        ).one_or_none()
+        if workflow_resource:
+            workflow_resource.quota_used = cpu_milliseconds
+        else:
+            workflow_resource = WorkflowResource(
+                workflow_id=workflow.id_,
+                resource_id=cpu_resource.id_,
+                quota_used=cpu_milliseconds,
+            )
+            user_resource_quota = UserResource.query.filter_by(
+                user_id=workflow.owner_id, resource_id=cpu_resource.id_
+            ).first()
+            user_resource_quota.quota_used += cpu_milliseconds
+            Session.add(workflow_resource)
+        Session.commit()
+        return cpu_milliseconds
+    return 0
+
+
+def update_users_cpu_quota(user=None) -> None:
+    """Update users CPU quota usage.
+
+    :param user: User whose CPU quota will be updated. If None, applies to all users.
+
+    :type user: reana_db.models.User
+
+    """
+    from reana_db.database import Session
+    from reana_db.models import (
+        ResourceType,
+        User,
+        UserResource,
+        UserToken,
+        UserTokenStatus,
+        Workflow,
+    )
+
+    if should_skip_quota_update(ResourceType.cpu):
+        return
+
+    if user:
+        users = [user]
+    else:
+        users = User.query.join(UserToken).filter_by(
+            status=UserTokenStatus.active  # skip users with no active token
+        )
+    for user in users:
+        cpu_milliseconds = 0
+        # logs and reana_specification are not loaded to avoid consuming
+        # huge amounts of memory
+        for workflow in user.workflows.options(
+            defer(Workflow.logs),
+            defer(Workflow.reana_specification),
+        ):
+            cpu_milliseconds += update_workflow_cpu_quota(workflow=workflow)
+        cpu_resource = get_default_quota_resource(ResourceType.cpu.name)
+        user_resource_quota = UserResource.query.filter_by(
+            user_id=user.id_, resource_id=cpu_resource.id_
+        ).first()
+        user_resource_quota.quota_used = cpu_milliseconds
+        Session.commit()
+
+
+def update_workspace_retention_rules(rules, status) -> None:
+    """Update workspace retention rules status.
+
+    :param rules: Workspace retention rules that need to be updated
+    :param status: Status accoring which retention rules need to be updated
+
+    :type rules: reana_db.models.WorkspaceRetentionRule
+    :type status: reana_db.models.WorkspaceRetentionRuleStatus
+    """
+    from reana_db.database import Session
+    from reana_db.models import WorkspaceRetentionRuleStatus
+
+    for rule in rules:
+        if rule.status == status:
+            continue
+        if not rule.can_transition_to(status):
+            raise Exception(
+                f"Cannot transition workspace retention rule {rule.id_} "
+                f"from status {rule.status} to {status}."
+            )
+        if status == WorkspaceRetentionRuleStatus.inactive:
+            rule.apply_on = None
+        if status == WorkspaceRetentionRuleStatus.active:
+            rule.apply_on = datetime.today().replace(
+                hour=23, minute=59, second=59
+            ) + timedelta(days=rule.retention_days)
+        rule.status = status
+        Session.add(rule)
+    Session.commit()
+
+
 def get_disk_usage_or_zero(workspace_path) -> int:
     """Get disk usage for the workspace if exists, zero if not."""
+    from reana_db.models import ResourceType
+
     if (
-        QuotaResourceType.disk not in WORKFLOW_TERMINATION_QUOTA_UPDATE_POLICY
-        and not CRONJOB_DISK_QUOTA_UPDATE_POLICY
+        ResourceType.disk.name not in WORKFLOW_TERMINATION_QUOTA_UPDATE_POLICY
+        and not PERIODIC_RESOURCE_QUOTA_UPDATE_POLICY
     ):
         return 0
 
-    from reana_commons.utils import get_disk_usage
-    from reana_commons.errors import REANAMissingWorkspaceError
-
     try:
         disk_bytes = get_disk_usage(workspace_path, summarize=True)
         return int(disk_bytes[0]["size"]["raw"])
@@ -276,7 +421,9 @@ def get_disk_usage_or_zero(workspace_path) -> int:
         return 0
 
 
-def store_workflow_disk_quota(workflow, bytes_to_sum: Optional[int] = None):
+def store_workflow_disk_quota(
+    workflow, bytes_to_sum: Optional[int] = None, override_policy_checks: bool = False
+):
     """
     Update or create disk workflow resource.
 
@@ -286,13 +433,16 @@ def store_workflow_disk_quota(workflow, bytes_to_sum: Optional[int] = None):
 
     :type workflow: reana_db.models.Workflow
     :type bytes_to_sum: int
-    """
-    if QuotaResourceType.disk not in WORKFLOW_TERMINATION_QUOTA_UPDATE_POLICY:
-        return
 
+    :param override_policy_checks: Whether to update the disk quota without checking the
+        update policy.
+    """
     from reana_db.database import Session
     from reana_db.models import ResourceType, WorkflowResource
 
+    if not override_policy_checks and should_skip_quota_update(ResourceType.disk):
+        return
+
     disk_resource = get_default_quota_resource(ResourceType.disk.name)
     workflow_resource = (
         Session.query(WorkflowResource)
@@ -301,8 +451,17 @@ def store_workflow_disk_quota(workflow, bytes_to_sum: Optional[int] = None):
     )
 
     if workflow_resource:
-        if bytes_to_sum:
-            workflow_resource.quota_used += bytes_to_sum
+        if bytes_to_sum is not None:
+            updated_quota_usage = workflow_resource.quota_used + bytes_to_sum
+            if updated_quota_usage < 0:
+                logging.warning(
+                    f"Disk quota consumption of workflow {workflow.id_} would become negative: "
+                    f"{workflow_resource.quota_used} [original usage] + {bytes_to_sum} [delta] "
+                    f"-> {updated_quota_usage} [new usage]. Setting the new usage to zero."
+                )
+                workflow_resource.quota_used = 0
+            else:
+                workflow_resource.quota_used = updated_quota_usage
         else:
             workflow_resource.quota_used = get_disk_usage_or_zero(
                 workflow.workspace_path
